#!/usr/bin/python
from lxml.objectify import NoneElement

__author__ = 'gabrielet'

# Import to handle mininet topology
from mininet.net import Mininet
from mininet.node import RemoteController, OVSSwitch, UserSwitch
from mininet.log import setLogLevel
from mininet.cli import CLI


# Modify this class if you want to use another switch
class CustomSwitch(OVSSwitch):

    def __init__(self, name, **params):
        OVSSwitch.__init__(self, name=name, datapath='user', **params)


# This class creates topology for conf/simple
class TopologyGenerator(object):

    def __init__(self):
        # If you want to use a UserSwitch
        self.net = Mininet(controller=None, switch=CustomSwitch, listenPort=6634, inNamespace=False, autoSetMacs=True)
        # If you want to use a OVSSwitch
        # self.net = Mininet(controller=None, switch=CustomSwitch)
        # Add controller to the network
        self.net.addController('c0', controller=RemoteController, ip='127.0.0.1', port=6633)

    # Function to generate an hex representation of dpid
    @staticmethod
    def dpid(dpid):
        return hex(dpid)[2:]

    def create_topology(self):
        # Core Switches
        pr1 = self.net.addSwitch("pr1", dpid=self.dpid(1))
        pr2 = self.net.addSwitch("pr2", dpid=self.dpid(2))
        pr3 = self.net.addSwitch("pr3", dpid=self.dpid(3))
        pr4 = self.net.addSwitch("pr4", dpid=self.dpid(4))
        pr5 = self.net.addSwitch("pr5", dpid=self.dpid(5))
        # PEs
        pe_mi = self.net.addSwitch("pe-mi", dpid=self.dpid(6))
        pe_rm = self.net.addSwitch("pe-rm", dpid=self.dpid(7))
        
        # Hosts
        h1 = self.net.addHost("h1")
        h2 = self.net.addHost("h2")

        # Hosts' links
        self.net.addLink(h1, pe_rm)
        self.net.addLink(h2, pe_mi)

        # Backbone links
        self.net.addLink(pr1, pr2)
        self.net.addLink(pr1, pr3)
        self.net.addLink(pr1, pr4)
        self.net.addLink(pr2, pr3)
        self.net.addLink(pr2, pr5)
        self.net.addLink(pr4, pr5)

        # PEs links
        self.net.addLink(pe_mi, pr1)
        self.net.addLink(pe_rm, pr5)

    def configure_host(self):
        ### Host h1
        h1 = self.net.get("h1")
        h1.setIP("10.0.0.1/24")
        # Add default route
        h1.cmd("route add default h1-eth0")
        # Add static ARP
        h1.cmd("arp -s 10.0.1.1 00:00:00:00:00:02")

        ### Host h2
        h2 = self.net.get("h2")
        h2.setIP("10.0.1.1/24")
        # Add default route
        h2.cmd("route add default h2-eth0")
        # Add static ARP
        h2.cmd("arp -s 10.0.0.1 00:00:00:00:00:01")

    def start_topology(self):
        # Start network
        self.net.start()
        # Configure hosts
        self.configure_host()
        # Open CLI
        CLI(self.net)
        self.net.stop()

if __name__ == "__main__":
    setLogLevel('info')
    t = TopologyGenerator()
    t.create_topology()
    t.start_topology()
