__author__ = 'gabrielet'

from netaddr import IPSet
from netaddr.ip import IPAddress, IPNetwork

from model.vpn import Datapath
from utils.pattern.singleton import Singleton

### Types of strategies
NAT_STATIC = "static"
NAT_DYNAMIC = "dynamic"
NAT_FULL = "full"

"""
This class models a NAT PE. A NAT PE is a datapath that performs NAT operations.
"""


class ProviderEdgeNat(Datapath):

    def __init__(self, dpid, name, label, outport):
        Datapath.__init__(self, dpid, name)
        # The label used to reach the natter
        self.label = label
        # The port used to forward the traffic in Internet
        self.outport = outport
        # This list contains a set of mapping that this NAT has to be handle
        self.mappings = []

    def add_mapping(self, mapping):
        self.mappings.append(mapping)

    def get_mapping_for_site(self, site):
        for mapping in self.mappings:
            if mapping.site is site:
                return mapping

    def install_rules(self, rule_generator, path):
        vpn_label = path.vpn.label
        next_hop = path.get_next_hop(self)
        # For each site of the VPN's path, take the mapping and install rules
        sites = path.vpn.get_all_sites()
        for site in sites:
            mapping = self.get_mapping_for_site(site)
            if mapping is not None:
        # For each mapping available on this NAT
        #for mapping in self.mappings:
                strategy = mapping.strategy
                dst_pe = mapping.site.pe
                site_ip = mapping.site.network
                ips = list(site_ip)
                for ip in ips:
                    gr = strategy.nat(ip)
                    # Check if gr is not None
                    if not gr is None:
                        # If NAT is source
                        if self is path.get_src_pe():
                            # Rewrite dst_ip on the packet
                            rule_generator.flow_mod_rewrite_public_dst_ip(
                                vpn_label=vpn_label, in_port=self.outport, dst_pe_label=dst_pe.label, next_hop=next_hop,
                                private_ip=ip, public_ip=gr, datapath=self.datapath, cookie=0, table_id=0,
                                idle_timeout=0, hard_timeout=0, priority=100)
                        # If NAT is destination
                        elif self is path.get_dst_pe():
                            # Remove VPN label
                            rule_generator.flow_mod_remove_vpn_at_nat(
                                vpn_label=vpn_label, datapath=self.datapath, cookie=0, table_id=0, idle_timeout=0,
                                hard_timeout=0, priority=100)
                            # Rewrite in accord with this mapping and forward
                            rule_generator.flow_mod_rewrite_private_src_ip(
                                vpn_label=vpn_label, private_ip=ip, public_ip=gr, outport=self.outport,
                                datapath=self.datapath, cookie=0, table_id=1, idle_timeout=0, hard_timeout=0,
                                priority=100)
        if self is not path.get_src_pe() and self is not path.get_dst_pe():
            # Install rules as a simple datapath
            super(ProviderEdgeNat, self).install_rules(rule_generator, path)

    def remove_rules(self, rule_generator, path):
        # Remove only for path in which this nat is source node
        src = path.get_src_pe()
        if self is src:
            # For each mapping of this NAT
            for mapping in self.mappings:
                strategy = mapping.strategy
                dst_pe = mapping.site.pe
                site_ip = mapping.site.network
                ips = list(site_ip)
                # For each rule of this strategy
                for ip in ips:
                    gr = strategy.nat(ip)
                    # Check if gr is not None
                    if not gr is None:
                        rule_generator.flow_mod_remove_rules_at_nat(
                            in_port=self.outport, public_ip=gr, datapath=self.datapath, cookie=0, table_id=0,
                            idle_timeout=0, hard_timeout=0, priority=100)
        elif self is not src and self is not path.get_dst_pe():
            # Remove rules as a simple datapath
            super(ProviderEdgeNat, self).remove_rules(rule_generator, path)


class Mapping(object):

    def __init__(self, site, strategy):
        # The site on which this mapping is applied to
        self.site = site
        # The NAT strategy (Static, Dynamic or Full)
        self.strategy = strategy

    def __repr__(self):
        #return "Mapping(" + repr(self.strategy) + ")"
        return "Mapping(site: %s, strategy=%s)" % (self.site, self.strategy)

# Factory class to create Strategy object (in accord with strategy pattern)
@Singleton
class FactoryStrategy:

    def __init__(self):
        self.strategy = None

    def create_strategy(self, strategy):
        if strategy == NAT_STATIC:
            self.strategy = StaticNat(NAT_STATIC)
        elif strategy == NAT_DYNAMIC:
            self.strategy = DynamicNat(NAT_DYNAMIC)
        elif strategy == NAT_FULL:
            self.strategy = FullNat(NAT_FULL)
        return self.strategy


class Strategy(object):

    def __init__(self, name):
        # The name of the natting strategy
        self.name = name
        # This map contains the association between private IP and public IP
        self.mapping = {}

    def __repr__(self):
        return "Strategy(" + self.name + ", " + repr(self.mapping) + ")"


class StaticNat(Strategy):

    def __init__(self, name):
        Strategy.__init__(self, name)

    def add_rule(self, ip, gr):
        self.mapping[IPAddress(ip)] = IPAddress(gr)

    def nat(self, ip):
        return self.mapping.get(ip, None)


class DynamicNat(Strategy):

    def __init__(self, name):
        Strategy.__init__(self, name)
        # The set of private IPs to map into GR IPs
        self.privateIPs = None
        # The starting pool of IP addresses. For each assigned IP, pop it form this list
        self.publicIPs = None

    def add_rule(self, ip, gr):
        self.privateIPs = list(IPNetwork(ip))
        self.publicIPs = list(IPNetwork(gr))
        # Set private and public IPs with the two different networks
        #self.privateIPs.add(IPNetwork(ip))
        #self.publicIPs.add(IPNetwork(gr))

    def nat(self, ip):
        gr = self.mapping.get(ip, None)
        # Check if IP is neither the network and broadcast address
        if ip.words[3] != 0 and ip.words[3] != 255:
            # Before creating a new mapping, try to look for an existing one
            if gr is None:
                if ip in self.privateIPs:
                    # Take a public IP
                    public_ip = self.publicIPs.pop(1)
                    print "Map", ip, "into", public_ip
                    # Take the private IP
                #self.privateIPs.remove(ip)
                # Take a public IP
                #list_gr = list(gr)
                #public_ip = list_gr.pop(0)
                #self.publicIPs.remove(public_ip)
                self.mapping[ip] = public_ip
        return gr


class FullNat(Strategy):

    def __init__(self, name):
        Strategy.__init__(self, name)
        # Available GlobalRealm IP addresses
        self.gr = IPSet()

    def add_rule(self, ip, gr):
        # Check if IP is "*"
        if ip == "*":
            self.gr.add(IPNetwork(gr))

    def nat(self, ip):
        # Check if a mapping already exists
        gr = self.mapping.get(ip, None)
        if gr is None:
            # The mapping does not exist, so take a public IP
            gr = self.gr.pop()
            # Create the mapping
            self.mapping[ip] = gr
        return gr
