__author__ = 'gabrielet'

from netaddr import *

"""
This class models a VPN object
"""


class VirtualPrivateNetwork(object):
    def __init__(self, name, label):
        # VPN's name
        self.name = name
        # Each VPN is identified by a label
        self.label = label
        # A VPN consists of a set of sites linked to the PEs in the network
        self.sites = []

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return "VPN(name=%s, label=%i)" % (self.name, self.label)

    def get_label(self):
        return self.label

    def add_site(self, site):
        self.sites.append(site)

    def get_all_sites(self):
        return self.sites

    def get_sites(self, pe):
        sites = []
        for site in self.sites:
            if site.vpn == self and site.pe == pe:
                sites.append(site)
        return sites

    def get_pes_on_ports(self):
        pes = {}
        # This method returns the map of <PE, list(port)> on which this VPN is linked
        for site in self.sites:
            # Take the port referred to this PE
            ports = pes.pop(site.pe, None)
            # If there is no PE in PEs, create a list of ports
            if ports is None:
                ports = [site.port]
            else:
                # PE is present: add port to the list
                ports.append(site.port)
            # Add the pair <PE, list(port)> into the map
            pes[site.pe] = ports
        return pes

    def get_networks(self):
        networks = []
        for site in self.sites:
            networks.append(site.network)
        return networks

    def info(self):
        print self
        for site in self.sites:
            print " * ", site
            mapping = site.natter.get_mapping_for_site(site)
            if mapping is not None:
                print "   - ", mapping
            else:
                print "   - This site cannot access on Internet!"


"""
This class models a site. Each VPN has one or more site linked on different PEs. So, a site represents the VPN subnet
linked at a certain PE at a well known port
"""


class Site(object):
    def __init__(self, vpn, pe, port, network):
        self.vpn = vpn
        self.pe = pe
        self.port = port
        self.network = IPNetwork(network)
        # The natter is initialize to None, because it will be added after at the site. It's a model.vpn.ProviderEdgeNat
        self.natter = None

    def __repr__(self):
        return "Site(VPN=%s, PE=%s, port=%s, network=%s, NAT=%s)" % (
            self.vpn.name, self.pe.name, self.port, self.network, repr(self.natter))

    def add_natter(self, nat):
        self.natter = nat

    def get_pe(self):
        return self.pe


"""
This class models a datapath. Simply, a datapath is an object identified by a name and a dpid;
moreover, a datapath wraps a ryu.controller.Datapath object
"""


class Datapath(object):
    def __init__(self, dpid, name):
        self.dpid = dpid
        self.name = name
        self.datapath = None

    def __hash__(self):
        return hash(self.dpid)

    def __repr__(self):
        return "Datapath(id=%s, name=%s)" % (self.dpid, self.name)

    # Set a datapath obj to this datapath; remember: datapath is ryu.controller.Datapath object!
    def set_datapath(self, datapath):
        self.datapath = datapath

    def install_rules(self, rule_generator, path):
        # Take dst PE (for routing label)
        dst_pe_label = path.get_dst_pe().label
        next_hop = path.get_next_hop(self)

        # Check if this node is penultimate hop
        if self is path.get_penultimate_hop():
            # Remove routing label and forward to next
            rule_generator.flow_mod_remove_routing_label_and_fwd(
                dst_pe_label=dst_pe_label, next_hop=next_hop, datapath=self.datapath, table_id=0, cookie=0,
                idle_timeout=0, hard_timeout=0, priority=100)
        else:
            # Forward to next hop
            rule_generator.flow_mod_forward(
                dst_pe_label=dst_pe_label, next_hop=next_hop, datapath=self.datapath, table_id=0, cookie=0,
                idle_timeout=0, hard_timeout=0, priority=100)

    def remove_rules(self, rule_generator, path):
        dst_pe_label = path.get_dst_pe().label
        # Remove rules based on dst label
        rule_generator.flow_mod_remove_rules_at_dp(
            dst_pe_label=dst_pe_label, datapath=self.datapath, table_id=0, cookie=0, idle_timeout=0, hard_timeout=0,
            priority=100)


"""
This class models a Provider Edge (PE)
"""


class ProviderEdge(Datapath):
    def __init__(self, dpid, name, label):
        # Superclass constructor
        Datapath.__init__(self, dpid, name)
        # A Provider Edge (PE) is an object that extends a Datapath. It also has a label, a
        # subnet and a port to which the subnet is attached
        self.label = label
        # A map of VPNs attached to this PE. VPN is VirtualPrivateNetwork instance.
        # map<port,vpn>, vpn instance of VirtualPrivateNetwork
        self.vpns = {}

    def add_vpn(self, port, vpn):
        self.vpns[port] = vpn

    def get_vpns(self):
        return self.vpns.values()

    def get_vpn_by_port(self, port):
        return self.vpns.get(port)

    def install_rules_for_local_delivery(self, rule_generator, vpn):
        print " - Install local delivery at PE", self.name, "belonging to VPN", vpn.name
        sites = vpn.get_sites(self)
        for src_site in sites:
            src_network = src_site.network
            in_port = src_site.port
            for dst_site in sites:
                if not src_site is dst_site:
                    outport = dst_site.port
                    dst_network = dst_site.network
                    rule_generator.flow_mod_add_local_delivery(
                        src_ip=src_network, in_port=in_port, dst_ip=dst_network, outport=outport,
                        datapath=self.datapath, cookie=0, table_id=0, idle_timeout=0, hard_timeout=0,
                        priority=100)

    def install_rules(self, rule_generator, path):
        # VPN label
        vpn_label = path.vpn.label
        # Take the src and dst PE
        src_pe = path.get_src_pe()
        dst_pe = path.get_dst_pe()
        # Next hop along the path
        next_hop = path.get_next_hop(self)
        # Take all sites at src and dst PEs
        src_sites = path.vpn.get_sites(src_pe)
        dst_sites = path.vpn.get_sites(dst_pe)

        # If this PE is src for given path
        if self is src_pe:
            # For all sites at src, install a rule to reach each site at dst PE. Scanning over all site at src PE is
            # needed because the match is based on in_port and dst_ip. In this way, produced rules are similar except
            # for in_port. If not so, isolation is not guarantee (other sites can reach sites in other VPN). The in_port
            # value guarantee isolation.
            for src_site in src_sites:
                in_port = int(src_site.port)
                src_ip = src_site.network
                # If src site has a natter, install rule
                if src_site.natter is not None:
                    nat_label = src_site.natter.label
                    # Install rule for reaching nat
                    rule_generator.flow_mod_add_vpn_and_dst_pe_label_and_fwd(
                        vpn_label=vpn_label, in_port=in_port, ip_src=src_ip, dst_pe_label=nat_label,
                        ip_dst=IPNetwork('0.0.0.0/0'), next_hop=next_hop, datapath=self.datapath, cookie=0,
                        table_id=0, idle_timeout=0, hard_timeout=0, priority=80)
                # Install a rule for each site at dst PE
                for dst_site in dst_sites:
                    dst_network = dst_site.network
                    # At dst_site take only site of the same src_site's VPN
                    if dst_site.vpn == src_site.vpn:

                        # Check if src PE (this) is the penultimate hop
                        if self is path.get_penultimate_hop():
                            # Add VPN label and forward
                            rule_generator.flow_mod_add_vpn_label_and_fwd(
                                vpn_label=vpn_label, in_port=in_port, ip_src=src_ip, ip_dst=dst_network,
                                next_hop=next_hop, datapath=self.datapath, cookie=0, table_id=0, idle_timeout=0,
                                hard_timeout=0, priority=100)
                        else:
                            # Add VPN and dst PE label and forward
                            rule_generator.flow_mod_add_vpn_and_dst_pe_label_and_fwd(
                                vpn_label=vpn_label, in_port=in_port, ip_src=src_ip, dst_pe_label=dst_pe.label,
                                ip_dst=dst_network, next_hop=next_hop, datapath=self.datapath, cookie=0, table_id=0,
                                idle_timeout=0, hard_timeout=0, priority=100)

        elif self is dst_pe:
            # For each site at dst, remove VPN label and forward to the destination
            for dst_site in dst_sites:
                dst_network = dst_site.network
                outport = dst_site.port
                # Remove VPN label
                rule_generator.flow_mod_remove_vpn_label(
                    vpn_label=vpn_label, datapath=self.datapath, table_id=0, cookie=0, idle_timeout=0,
                    hard_timeout=0, priority=100)
                # Forward to destination
                rule_generator.flow_mod_forward_to_dst(
                    vpn_label=vpn_label, dst_ip=dst_network, outport=outport, datapath=self.datapath, cookie=0,
                    table_id=1, idle_timeout=0, hard_timeout=0, priority=100)
        else:
            # Install rule as simple datapath
            super(ProviderEdge, self).install_rules(rule_generator, path)

    def remove_rules(self, rule_generator, path):
        # At a PE, the rules to remove have the match (in_port + src_ip + dst_ip)
        # VPN label
        vpn_label = path.vpn.label
        # Take the src and dst PE
        src_pe = path.get_src_pe()
        dst_pe = path.get_dst_pe()
        # Take all sites at src and dst PEs
        src_sites = path.vpn.get_sites(src_pe)
        dst_sites = path.vpn.get_sites(dst_pe)

        # Do something only if this PE is the src
        if self is src_pe:
            # Scan src_sites
            for src_site in src_sites:
                in_port = src_site.port
                src_ip = src_site.network

                # Check if destination is a natter for this site
                if dst_pe == src_site.natter:
                    rule_generator.flow_mod_remove_rules_at_pe(
                        in_port=in_port, ip_src=src_ip, ip_dst=IPNetwork('0.0.0.0/0'), datapath=self.datapath, cookie=0,
                        table_id=0, idle_timeout=0, hard_timeout=0, priority=80)

                for dst_site in dst_sites:
                    dst_ip = dst_site.network
                    rule_generator.flow_mod_remove_rules_at_pe(
                        in_port=in_port, ip_src=src_ip, ip_dst=dst_ip, datapath=self.datapath, cookie=0, table_id=0,
                        idle_timeout=0, hard_timeout=0, priority=100)
        elif self is not src_pe and self is not dst_pe:
            # Remove rules as simple datapath
            super(ProviderEdge, self).remove_rules(rule_generator, path)