__author__ = 'gabriele'

from ryu.ofproto import ether

"""
This class implements a rules installer. It invokes the rule generator for taking the correct rules; then it sends the
rules to all node in the path. In this package is wrapped the API of the used controller (e.g. Ryu)
"""


class RuleInstaller(object):
    # topology is the parameter useful to take outport of datapath for traffic forward
    def __init__(self, topology):
        # The topology useful to retrieve the outport
        self.topology = topology
        # Rule generator
        self.rules_generator = RuleGenerator(topology)

    def send_flow_mods_for_local_delivery(self, vpn, pe):
        pe.install_rules_for_local_delivery(self.rules_generator, vpn)

    def send_flow_mods(self, path):
        # Take src, dps and dst
        src_pe = path.get_src_pe()
        dps = path.get_intermediate_dps()
        dst_pe = path.get_dst_pe()

        # Install rules on path
        src_pe.install_rules(self.rules_generator, path)

        # Install rules for routing traffic at intermediate datapaths
        for dp in dps:
            # Take into account the src_pe_label, useful for replay
            dp.install_rules(self.rules_generator, path)

        dst_pe.install_rules(self.rules_generator, path)

        # Take reverse path
        r_path = path.create_reverse_path()
        # Take src, dps and dst
        r_src_pe = r_path.get_src_pe()
        r_dps = r_path.get_intermediate_dps()
        r_dst_pe = r_path.get_dst_pe()

        # Install rules on reverse path
        r_src_pe.install_rules(self.rules_generator, r_path)

        # Install rules for routing traffic at intermediate datapaths
        for r_dp in r_dps:
            # Take into account the src_pe_label, useful for replay
            r_dp.install_rules(self.rules_generator, r_path)

        r_dst_pe.install_rules(self.rules_generator, r_path)

    def send_flow_rems(self, path):
        # Take src, dps and dst
        src_pe = path.get_src_pe()
        dps = path.get_intermediate_dps()
        dst_pe = path.get_dst_pe()

        # Install rules on path
        src_pe.remove_rules(self.rules_generator, path)

        # Install rules for routing traffic at intermediate datapaths
        for dp in dps:
            # Take into account the src_pe_label, useful for replay
            dp.remove_rules(self.rules_generator, path)

        dst_pe.remove_rules(self.rules_generator, path)

        # Take reverse path
        r_path = path.create_reverse_path()
        # Take src, dps and dst
        r_src_pe = r_path.get_src_pe()
        r_dps = r_path.get_intermediate_dps()
        r_dst_pe = r_path.get_dst_pe()

        # Install rules on reverse path
        r_src_pe.remove_rules(self.rules_generator, r_path)

        # Install rules for routing traffic at intermediate datapaths
        for r_dp in r_dps:
            # Take into account the src_pe_label, useful for replay
            r_dp.remove_rules(self.rules_generator, r_path)

        r_dst_pe.remove_rules(self.rules_generator, r_path)


"""
This class implements a rules generator. It wraps the Ryu API for rules management and generates the rules to install
over the path.
"""


class RuleGenerator(object):
    def __init__(self, topology):
        # From topology rules generator may take the outport of datapath for traffic forwarding
        self.topology = topology
        self.cookie_mask = 0
        self.buffer_id = None
        self.importance = 0

        # Offset for VLAN matching: values in the range 1-4095 are not valid. So, an offset of 4096 is needed
        self._vlan_offset = 4096
        # Reference count for flow modifications (useful when a rule has to be deleted)
        # This map have the form <OFPMatch, count>
        self._ref_counting = {}

    def _update_ref_count(self, match):
        # Using a string representation of the match guarantees equal hash for equal match
        str_match = str(match)
        count = self._ref_counting.get(str_match, 0)
        count += 1
        self._ref_counting[str_match] = count

    def _rule_can_be_removed(self, match):
        may_remove = False
        str_match = str(match)
        count = self._ref_counting.get(str_match)
        if count == 1:
            may_remove = True
        return may_remove

    def _decrease_ref_count(self, match):
        str_match = str(match)
        count = self._ref_counting.get(str_match)
        if count == 1:
            del self._ref_counting[str_match]
        else:
            count -= 1
            self._ref_counting[str_match] = count

    def get_ref_counting(self):
        return self._ref_counting

    ### In each method, datapath parameter is an instance of ryu.controller.Datapath ###

    def flow_mod_add_local_delivery(self, src_ip, in_port, dst_ip, outport,
                                    datapath, cookie, table_id, idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(in_port=int(in_port), eth_type=ether.ETH_TYPE_IP,
                                    ipv4_src=(src_ip.network, src_ip.netmask),
                                    ipv4_dst=(dst_ip.network, dst_ip.netmask))
        self._update_ref_count(match)

        # Action
        actions = [
            ofp_parser.OFPActionOutput(int(outport))  # Outport
        ]
        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions)
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_add_vpn_label_and_fwd(self, vpn_label, in_port, ip_src, ip_dst, next_hop,
                                       datapath, cookie, table_id, idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(
            in_port=in_port, eth_type=ether.ETH_TYPE_IP, ipv4_src=(ip_src.network, ip_src.netmask),
            ipv4_dst=(ip_dst.network, ip_dst.netmask))
        self._update_ref_count(match)

        # Action
        outport = self.topology.get_outport(datapath.id, next_hop.dpid)
        actions = [
            # Add VPN label
            ofp_parser.OFPActionPushVlan(),
            # VLAN VID until 4095 are not valid; so add an offset of 4096
            ofp_parser.OFPActionSetField(vlan_vid=vpn_label + self._vlan_offset),
            # Forward
            ofp_parser.OFPActionOutput(int(outport))
        ]
        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions)
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_add_vpn_and_dst_pe_label_and_fwd(
            self, vpn_label, in_port, ip_src, dst_pe_label, ip_dst, next_hop,
            datapath, cookie, table_id, idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(
            in_port=in_port, eth_type=ether.ETH_TYPE_IP, ipv4_src=(ip_src.network, ip_src.netmask),
            ipv4_dst=(ip_dst.network, ip_dst.netmask))
        self._update_ref_count(match)

        # Action
        outport = self.topology.get_outport(datapath.id, next_hop.dpid)
        actions = [
            # Add VPN label
            ofp_parser.OFPActionPushVlan(),
            # VLAN VID until 4095 are not valid; so add an offset of 4096
            ofp_parser.OFPActionSetField(vlan_vid=vpn_label + self._vlan_offset),
            # Add dst Pe label
            ofp_parser.OFPActionPushMpls(),
            ofp_parser.OFPActionSetField(mpls_label=dst_pe_label),
            # Output
            ofp_parser.OFPActionOutput(int(outport))
        ]
        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions)
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_forward(self, dst_pe_label, next_hop, datapath, table_id, cookie, idle_timeout, hard_timeout,
                         priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Match structure
        match = ofp_parser.OFPMatch(eth_type=ether.ETH_TYPE_MPLS, mpls_label=dst_pe_label)
        self._update_ref_count(match)

        # Actions
        outport = self.topology.get_outport(datapath.id, next_hop.dpid)
        actions = [
            ofp_parser.OFPActionOutput(int(outport))  # Outport
        ]

        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions)
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_remove_routing_label_and_fwd(
            self, dst_pe_label, next_hop, datapath, table_id, cookie, idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Match structure
        match = ofp_parser.OFPMatch(eth_type=ether.ETH_TYPE_MPLS, mpls_label=dst_pe_label)
        self._update_ref_count(match)

        # Actions
        outport = self.topology.get_outport(datapath.id, next_hop.dpid)
        actions = [
            # Add routing label by:
            ofp_parser.OFPActionPopMpls(),
            ofp_parser.OFPActionOutput(int(outport))
        ]

        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions)
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_remove_vpn_label(self, vpn_label, datapath, table_id, cookie, idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Match structure
        match = ofp_parser.OFPMatch(vlan_vid=vpn_label + self._vlan_offset)
        self._update_ref_count(match)

        # Actions
        actions = [
            # Remove VPN label
            ofp_parser.OFPActionPopVlan()
        ]

        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions),
            ofp_parser.OFPInstructionWriteMetadata(vpn_label, 0xffffff),
            ofp_parser.OFPInstructionGotoTable(1)
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_forward_to_dst(
            self, vpn_label, dst_ip, outport, datapath, cookie, table_id, idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(metadata=vpn_label, eth_type=ether.ETH_TYPE_IP,
                                    ipv4_dst=(dst_ip.network, dst_ip.netmask))
        self._update_ref_count(match)

        # Actions
        actions = [
            ofp_parser.OFPActionOutput(int(outport))  # Outport
        ]

        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions)
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_rewrite_public_dst_ip(
            self, vpn_label, in_port, dst_pe_label, next_hop, private_ip, public_ip, datapath, cookie, table_id,
            idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(in_port=int(in_port), eth_type=ether.ETH_TYPE_IP, ipv4_dst=public_ip)
        self._update_ref_count(match)

        # Actions
        outport = self.topology.get_outport(datapath.id, next_hop.dpid)
        actions = [
            ofp_parser.OFPActionSetField(ipv4_dst=str(private_ip)),
            ofp_parser.OFPActionPushVlan(),
            ofp_parser.OFPActionSetField(vlan_vid=vpn_label + self._vlan_offset),
            ofp_parser.OFPActionPushMpls(),
            ofp_parser.OFPActionSetField(mpls_label=dst_pe_label),
            ofp_parser.OFPActionOutput(int(outport))
        ]

        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions),
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_remove_vpn_at_nat(self, vpn_label, datapath, cookie, table_id, idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(vlan_vid=vpn_label + self._vlan_offset)
        self._update_ref_count(match)

        # Actions
        actions = [
            ofp_parser.OFPActionPopVlan(),
            #ofp_parser.OFPActionSetField(ipv4_src=public_ip)
            #ofp_parser.OFPActionOutput(int(outport))
        ]

        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions),
            ofp_parser.OFPInstructionWriteMetadata(vpn_label, 0xffffff),
            ofp_parser.OFPInstructionGotoTable(1)
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_rewrite_private_src_ip(
            self, vpn_label, private_ip, public_ip, outport, datapath, cookie, table_id, idle_timeout, hard_timeout,
            priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(metadata=vpn_label, eth_type=ether.ETH_TYPE_IP, ipv4_src=private_ip)
        self._update_ref_count(match)

        # Actions
        actions = [
            ofp_parser.OFPActionSetField(ipv4_src=public_ip),
            ofp_parser.OFPActionOutput(int(outport))
        ]

        # Instructions set
        inst = [
            ofp_parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions)
        ]

        # Flow modification
        flow_mod = ofp_parser.OFPFlowMod(
            datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_ADD, idle_timeout, hard_timeout, priority,
            self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
        datapath.send_msg(flow_mod)

    def flow_mod_remove_rules_at_pe(self, in_port, ip_src, ip_dst, datapath, cookie, table_id, idle_timeout,
                                    hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(
            in_port=int(in_port), eth_type=ether.ETH_TYPE_IP, ipv4_src=(ip_src.network, ip_src.netmask),
            ipv4_dst=(ip_dst.network, ip_dst.netmask))

        inst = []

        # Check ref counting to remove this rule
        if self._rule_can_be_removed(match):
            # Flow modification
            flow_mod = ofp_parser.OFPFlowMod(
                datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_DELETE_STRICT, idle_timeout, hard_timeout,
                priority, self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
            datapath.send_msg(flow_mod)
        # However, decrease the ref counting
        self._decrease_ref_count(match)

    def flow_mod_remove_rules_at_dp(
            self, dst_pe_label, datapath, table_id, cookie, idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(eth_type=ether.ETH_TYPE_MPLS, mpls_label=dst_pe_label)

        inst = []

        # Check ref counting to remove this rule
        if self._rule_can_be_removed(match):
            # Flow modification
            flow_mod = ofp_parser.OFPFlowMod(
                datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_DELETE_STRICT, idle_timeout, hard_timeout,
                priority, self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
            datapath.send_msg(flow_mod)
        # However, decrease the ref counting
        self._decrease_ref_count(match)

    def flow_mod_remove_rules_at_nat(
            self, in_port, public_ip, datapath, cookie, table_id, idle_timeout, hard_timeout, priority):
        # By default cookie_mask=cookie
        self.cookie_mask = cookie
        ofp = datapath.ofproto
        ofp_parser = datapath.ofproto_parser

        # Setting buffer
        self.buffer_id = ofp.OFP_NO_BUFFER

        # Create rule
        match = ofp_parser.OFPMatch(in_port=int(in_port), eth_type=ether.ETH_TYPE_IP, ipv4_dst=public_ip)

        inst = []

        # Check ref counting to remove this rule
        if self._rule_can_be_removed(match):
            # Flow modification
            flow_mod = ofp_parser.OFPFlowMod(
                datapath, cookie, self.cookie_mask, table_id, ofp.OFPFC_DELETE_STRICT, idle_timeout, hard_timeout,
                priority, self.buffer_id, ofp.OFPP_ANY, ofp.OFPG_ANY, ofp.OFPFF_SEND_FLOW_REM, match, inst)
            datapath.send_msg(flow_mod)
        # However, decrease the ref counting
        self._decrease_ref_count(match)