__author__ = 'gabrielet'

import os
import ConfigParser
import itertools

from time import *

from ofp.rule import RuleInstaller
from ofp.command import CMD_ADD, CMD_REM
from topology.topology import Topology
from topology.search import PathSearcher
from utils.parser import Parser

SYS_CONFIG = "conf/system.conf"

"""
This class represents the system (in accord with Controller pattern)
"""


class VpnSystem(object):

    def __init__(self):
        # Root directory of the controller
        self.sys_root = os.path.dirname(os.path.abspath(__file__))
        # Load system configuration
        self.system_conf = self.sys_root + "/" + SYS_CONFIG
        # System config parser
        self.system_config_parser = ConfigParser.ConfigParser()

        ### Initialize internal data structures
        # The list of VPNs
        self.vpns = {}  # Each object is a VirtualPrivateNetworks instance
        # The map of datapaths
        self.dps = {}
        # The map of PEs <dpid, ProviderEdge>
        self.pes = {}
        # The map of policies. <vpn, Policy>
        self.policies = {}
        # A map with list of PEs' pair of this VPN; useful to perform paths' search. This map has the form
        # <VPN.name, [[p1,p2], [p1,p3], ...]. The key is the name of the VPN otherwise the value is a list; each
        # element of this list is, in turn, a list of size 2. The former list contains all pairs of PEs for a VPN;
        # on the other hand, the latter list contains a pair of PE
        self.pes_pair = {}
        # A map of <VPN,list(PE)> with local delivery rules
        self.pes_with_local_delivery = {}

        # Create parser object
        self.parser = Parser()
        # Create a topology instance
        self.topology = Topology()
        # Path searcher
        self.path_searcher = None
        # Rules installer
        self.rules_installer = RuleInstaller(self.topology)

    def load_system_configuration(self):
        # Create a ConfigParser to load system configuration
        self.system_config_parser.readfp(open(self.system_conf))
        # Load VPNs configuration file
        vpns_conf = self.sys_root + "/" + self.system_config_parser.get('System', 'vpn-config-file')
        # Start parser
        self.parser.load(vpns_conf)

    def init(self):
        # Load from parser the maps of VPNs and datapaths
        self.vpns = self.parser.get_vpns()
        self.dps = self.parser.get_dps()
        
        #for vpn in self.vpns.values():
        #    print vpn.info()

        # Compute the list of all PEs' pairs for each VPN (useful when a path has to be searched)
        self._compute_pes_pairs()

        # Load policies
        for vpn in self.vpns.keys():
            policy = self.system_config_parser.get('Policies', vpn)
            key = self.vpns.get(vpn)    # key is a model.vpn.VirtualPrivateNetwork object
            self.policies[key] = policy

        # Create and start path searcher
        self.path_searcher = PathSearcher(self.vpns, self.policies, self.pes_pair)
        # Add this object as Observer for PathSearcher
        self.path_searcher.add_observer(self)
        self.path_searcher.start()

        # Install rules for local delivery
        for vpn in self.vpns.values():
            pes = []    # A list of PE with more than 1 site for vpn
            pes_on_port = vpn.get_pes_on_ports()
            for pe in pes_on_port.keys():
                sites = pes_on_port.get(pe, None)
                if sites is not None and len(sites) >= 2:
                    pes.append(pe)
            self.pes_with_local_delivery[vpn] = pes

    # This method computes all PEs' pairs for each VPN
    def _compute_pes_pairs(self):
        # For each VPN, calculate the list of all PE's pairs
        # In this for cycle, calculate all possible combinations of PEs for each VPN
        for vpn in self.vpns.values():
            pes = []
            for site in vpn.sites:
                if not site.pe in pes:
                    pes.append(site.pe)
                pairs = itertools.combinations(pes, 2)
            self.pes_pair[vpn] = pairs

        # In this cycle, convert a combinations object into a list of lists.
        for vpn in self.pes_pair.keys():
            pairs = self.pes_pair.pop(vpn)
            pes = []    # The list that contains all lists that represent the PEs' pairs
            for pair in pairs:
                pes.append([pair[0], pair[1]])  # Each item of pes list is a list of only 2 elements (the PE's pair)
            self.pes_pair[vpn] = pes

    # This method is implemented in accord with Observer pattern
    def update(self, command, path):
        # Check command parameter for detecting if install or remove rule
        if command == CMD_ADD:
            print "New", path, "found!"
            # Install rule
            self.rules_installer.send_flow_mods(path)
        elif command == CMD_REM:
            # Remove rule
            print "Removing rules for path", repr(path)
            self.rules_installer.send_flow_rems(path)
            # Search new path
            self.path_searcher.search_single_path(self.topology, path.vpn, path.get_src_pe(), path.get_dst_pe())

    # Map interface name into port number
    def map(self, datapath):
        ports = datapath.ports

        # For each VPN, check if this datapath is a PE for its and map interface name into port number
        for vpn in self.vpns.values():
            sites = vpn.get_all_sites()
            for site in sites:
                if site.pe.dpid == datapath.id:
                    site_port = site.port
                    # Find the right datapath's port
                    for port in ports.values():
                        if site_port in port.name:
                            site.port = int(port.port_no)
                # If this datapath is also natter
                if site.natter is not None and site.natter.dpid == datapath.id:
                    natter_port = site.natter.outport
                    for port in ports.values():
                        if natter_port in port.name:
                            site.natter.outport = int(port.port_no)

    # Add a new node to the topology. datapath is a ryu.controller.Datapath object, and it is contained into
    # event object of switch_enter_handler method
    def add_node(self, datapath):
        # When a datapath (ryu.controller.Datapath instance) arrives, set it into model.vpn.Datapath object
        # and add it to the topology

        dp = self.dps.pop(datapath.id, None)    # dp is model.vpn.{Datapath,ProviderEdge,ProviderEdgeNat} instance
        if not dp is None:
            # Set the ryu.controller.Datapath instance (useful for further switch/controller communication)
            dp.set_datapath(datapath)
            self.dps[dp.dpid] = dp

        # Now, add dp to the topology
        self.topology.add_node(dp)

        # Install rules for local delivery, if dp is a PE and it has more than one site for a given VPN
        for vpn in self.pes_with_local_delivery.keys():
            pes = self.pes_with_local_delivery.get(vpn, None)
            if dp in pes:
                self.rules_installer.send_flow_mods_for_local_delivery(vpn, dp)

    # Delete a node from the topology. datapath is a ryu.controller.Datapath object, and it is contained into
    # event object of switch_leave_handler method
    def del_node(self, datapath):
        pass

    # Add a new link to the topology. link is a ryu.controller.Link (?) object, and it is contained into
    # event object of link_add_handler method
    def add_link(self, link):
        # When a new link is detected (through LLDP protocol), add it to the topology
        self.topology.add_link(link)

    # Delete a link from the topology. link is a ryu.controller.Link (?) object, and it is contained into
    # event object of link_delete_handler method
    def delete_link(self, link):
        self.topology.delete_link(link)

        # Take src and dst (they are integer value!)
        src = link.src.dpid
        dst = link.dst.dpid
        # Check if those nodes are included in some path
        self.path_searcher.check_if_link_is_in_a_path(src, dst)

    # Search paths between a pair of PEs of the same VPN
    def search(self):
        # Update topology in search_path
        self.path_searcher.search(self.topology)

    def info(self):
        for vpn in self.vpns.values():
            vpn.info()
