__author__ = 'gabriele'

from threading import Thread
import networkx as nx
from networkx import NetworkXNoPath

from utils.pattern.observer import Observable
from ofp.command import CMD_ADD, CMD_REM

"""
This class models a path. Simply, a path is a list of nodes, a src PE, a dst PE and the VPN which PEs belong to.
It has also in charge the task of installing rules at the nodes included in the path.
"""


class Path(object):

    def __init__(self):
        pass

    # vpn is a model.vpn.VirtualPrivateNetwork object
    # src and dst are model.vpn.ProviderEdge objects
    def __init__(self, vpn, nodes):
        # The VPN to which this path belongs to
        self.vpn = vpn

        # Source PE; is a graph.Node object. It wraps a PE (model.vpn.ProviderEdge) object
        self.src = nodes.pop(0)     # The src
        # Destination PE; is a graph.Node object. It wraps a PE (model.vpn.ProviderEdge) object
        self.dst = nodes.pop(len(nodes) - 1)    # The dst

        # List of graph.Node objects; they wrap a datapath (model.vpn.Datapath) object
        self.nodes = nodes

    def __repr__(self):
        path = ""
        if self.len() == 1:
            path = "Path[" + self.vpn.name + "](" + self.src.datapath.name + ")"
        else:
            path = "Path[" + self.vpn.name + "](" + self.src.datapath.name + " -- "
            for n in self.nodes:
                path += n.datapath.name + " -- "
            path += self.dst.datapath.name + ")"
        return path

    # Create a reverse path useful for going back rules
    def create_reverse_path(self):
        nodes = list(self.nodes)
        nodes.reverse()
        # Add src and dst to the nodes list
        nodes.insert(0, self.dst)
        nodes.insert(len(nodes), self.src)
        r_path = Path(self.vpn, nodes)
        return r_path

    def len(self):
        if self.src == self.dst:
            path_len = 1
        elif len(self.nodes) == 0:
            path_len = 2
        else:
            path_len = len(self.nodes) + 2
        return path_len

    # Return the src PE (model.vpn.ProviderEdge)
    def get_src_pe(self):
        return self.src.datapath

    # Return the list of intermediate datapaths (model.vpn.Datapath)
    def get_intermediate_dps(self):
        dps = []
        for node in self.nodes:
            dps.append(node.datapath)
        return dps

    # Return the dst PE (model.vpn.ProviderEdge)
    def get_dst_pe(self):
        return self.dst.datapath

    # Remember that node is a model.vpn.{Datapath,ProviderEdge} instance, whereas src, dst and nodes are graph.Node
    def get_next_hop(self, node):
        next_hop = None
        # If node is src, then the next hop is the first element of the list
        if node == self.src.datapath:
            next_hop = self.nodes[0].datapath
        # Else, look for next hop in intermediate nodes
        else:
            count = 0
            found = False
            # Looking for this node
            for n in self.nodes:
                # If found it and this is not the last node in the list
                if not found:
                    if node == n.datapath and count < len(self.nodes) - 1:
                        # The next hop is the neighbor of "node"
                        next_hop = self.nodes[count + 1].datapath
                        found = True
                    else:
                        # If the "node" is the last in the intermediate nodes, the the next hop is dst node in the path
                        next_hop = self.dst.datapath
                count += 1
        return next_hop

    # Return the penultimate hop (model.vpn.{Datapath,ProviderEdge) in the path
    def get_penultimate_hop(self):
        if len(self.nodes) == 0:
            penultimate_hop = self.src.datapath
        else:
            penultimate_hop = self.nodes[len(self.nodes) - 1].datapath
        return penultimate_hop

    # Check if this path contains src and/or dst
    def contains(self, src, dst):
        contains = False
        count = 0
        # Path of len 2: check if src is self.src and dst is self.dst
        if len(self.nodes) == 0:
            if self.src.dpid == src and self.dst.dpid == dst:
                contains = True
        else:
            if self.src.dpid == src and self.nodes[0].dpid == dst:
                contains = True
            else:
                for node in self.nodes:
                    if count < len(self.nodes) - 1:
                        if node.dpid == src and self.nodes[count + 1].dpid == dst:
                            contains = True
                    else:
                        if node.dpid == src and self.dst.dpid == dst:
                            contains = True
                    count += 1
        return contains

"""
This class implements the search strategy in the graph. This is a thread, and it starts at VpnSystem constructor.
"""


class PathSearcher(Thread, Observable):

    # network is an instance of topology.topology.Topology
    def __init__(self, vpns, policies, pes_pairs):
        Thread.__init__(self)
        Observable.__init__(self)
        # All paths computed by all policies; it is a map in the form map<vpn, list of paths>. The key is a string (vpn
        # name)
        self.all_paths = {}

        # Tha map of all VPN in the system
        self.vpns = vpns
        # The map of all pair of PEs per VPN
        self.pes_pairs = pes_pairs
        # Tha map of the policies
        self.policies = policies
        # This map maintains the association between a VPN and its policy search. <vpn, PolicyPathSearch>
        self.searchers = {}

    # This method is implemented in accord with Observer pattern
    def update(self, command, path):
        ### When PathSearcher is updated by an observable, add the path to the map of all paths
        # Take the paths of the VPN
        paths = self.all_paths.pop(path.vpn, None)
        if paths is None:
            # Create a paths list for path.vpn and add to the path to the list
            paths = [path]
        else:
            paths.append(path)
        # Add the lists to the map of all paths
        self.all_paths[path.vpn] = paths
        # Now, notify all observers for this object (namely VpnSystem)
        self.notify_all(command, path)

    # Topology is changed due to add/remove of link or node
    def search(self, topology):
        # If topology changes, then start a new search on each for each vpn
        for vpn in self.pes_pairs.keys():
            # Take the correspondent searcher
            searcher = self.searchers.get(vpn)
            # For each PEs' pair, then search a path
            pes_pairs = self.pes_pairs.get(vpn)
            for pair in pes_pairs:
                src = topology.get_node(pair[0].dpid)
                dst = topology.get_node(pair[1].dpid)
                # Search only if src and dst are already present in the nx_graph
                if topology.contain(src) and topology.contain(dst):
                    searcher.search(topology, vpn, src, dst)

        # Search a path between a PE and a NAT PE (if the configuration allows)
        for vpn in self.vpns.values():
            # Looking for a site with a natter not equal to None
            sites = vpn.get_all_sites()
            for site in sites:
                if site.natter is not None:
                    searcher = self.searchers.get(vpn)
                    src = topology.get_node(site.pe.dpid)
                    dst = topology.get_node(site.natter.dpid)
                    if topology.contain(src) and topology.contain(dst):
                        searcher.search(topology, vpn, src, dst)

    def search_single_path(self, topology, vpn, src, dst):
        # Search a path for a VPN between src and dst
        searcher = self.searchers.get(vpn)
        graph_src = topology.get_node(src.dpid)
        graph_dst = topology.get_node(dst.dpid)
        searcher.search(topology, vpn, graph_src, graph_dst)

    # Check if a link from src to dst is included into a path. If so, notify all observers
    def check_if_link_is_in_a_path(self, src, dst):
        # Scan path map for vpn
        for vpn in self.all_paths.keys():
            paths = self.all_paths.get(vpn)
            for path in paths:
                # Check if path contains src and/or dst
                if path.contains(src, dst):
                    # Remove from paths
                    paths.remove(path)
                    # Remove from searcher
                    searcher = self.searchers.get(vpn)
                    searcher.remove_path_from_list(path)
                    self.notify_all(CMD_REM, path)

    def run(self):
        print "Path searcher started! Create policy objects and keep an association with the right VPN..."
        # For each VPN policy, create an instance of PolicyPathSearch and add into searchers map
        for vpn in self.policies.keys():
            # Take the policy
            policy = self.policies.pop(vpn)
            policy_path_search = eval(policy)()
            # To each policy_path_server add this object as observer
            policy_path_search.add_observer(self)
            # In this way, each VPN has its own policy_path_searcher
            self.searchers[vpn] = policy_path_search


"""
This is an abstract class that model a policy to search one (or more) path(s)
"""


class PolicyPathSearch(Observable):

    def __init__(self):
        Observable.__init__(self)
        # The vpn on which this policy runs
        self.vpn = None
        # The list of paths; each element is a Path object
        self.paths = []


class ShortestPath(PolicyPathSearch):

    def __init__(self):
        PolicyPathSearch.__init__(self)

    def search(self, topology, vpn, src, dst):
        # Set VPN
        self.vpn = vpn
        try:
            nx_path = nx.shortest_path(topology.nx_graph, src, dst)
            # Create a Path object
            path = Path(self.vpn, nx_path)

            if not self._exist_path_between_src_and_dst(path):
                # Add path
                self.paths.append(path)
                # Notify observers
                self.notify_all(CMD_ADD, path)
        except NetworkXNoPath:
            pass

    def _exist_path_between_src_and_dst(self, path):
        found = False
        # Scan paths list looking for a path with same path.src and path.dst
        for p in self.paths:
            if not found and p.src == path.src and p.dst == path.dst:
                found = True
        return found

    def remove_path_from_list(self, path):
        self.paths.remove(path)