__author__ = 'gabrielet'

import networkx as nx
from networkx import NetworkXError

from graph import Node, OutportGraph

"""
This class models the topology; in particular, a topology object has in charge the task of
maintaining updated two types of graphs: one for the path search algorithms and another for
mapping of outports used in flow modification rules.
"""


class Topology(object):

    def __init__(self):
        # Networkx graph (for path search)
        self.nx_graph = nx.Graph()
        # Outport graph (for ports to use in flow modification rules)
        self.outport_graph = OutportGraph()
        # The map of already added nodes. This map is useful when a new link arrive,
        # to take a Node instance starting from dpid of datapath in Link object. Otherwise,
        # when I add a link to nx_graph, add a node represented by a dpid instead of a Node
        # instance. Very important!!! Not delete this map!!!
        self.nodes = {}  # <node.dpid, topology.graph.Node>

    # datapath is Datapath (model.vpn.Datapath) object
    def add_node(self, datapath):
        # When a new node arrives, create and add it to both graphs
        # Each node wraps a datapath object (model.vpn.Datapath)
        node = Node(datapath.dpid, datapath)
        self.nodes[node.dpid] = node
        # Add node to the graph
        self.nx_graph.add_node(node)
        self.outport_graph.add_node(node)

    # A datapath fails for some reason
    def delete_node(self, datapath):
        pass

    # link is a Link (ryu.topology.switches.Link)
    def add_link(self, link):
        src = self.nodes.get(link.src.dpid)
        dst = self.nodes.get(link.dst.dpid)
        self.nx_graph.add_edge(src, dst)
        # Make the graph undirected: when a (direct) link arrives, in outport add
        self.outport_graph.add_link(src, dst, link.src.port_no)
        self.outport_graph.add_link(dst, src, link.dst.port_no)

    # A link fails for some reason
    def delete_link(self, link):
        # Remove edge form nx_graph
        src = self.nodes.get(link.src.dpid)
        dst = self.nodes.get(link.dst.dpid)

        try:
            self.nx_graph.remove_edge(src, dst)
        except NetworkXError:   # Between src and dst arrive 2 edges: src->dst and dst->src. xn_graph is undirected
                                # so the second link raises a NetworkXError (see remove_edge docs).
                                # In that case do nothing.
            pass

        # Remove edge from outport_graph
        self.outport_graph.remove_edge(src, dst)
        self.outport_graph.remove_edge(dst, src)

    def contain(self, node):
        if node in self.nx_graph:
            return True
        else:
            return False

    # Return a graph.Node object starting from its dpid
    def get_node(self, dpid):
        return self.nodes.get(dpid)

    # This method return the outport of node src_node used to reach dst_node. src_node and dst_node are int (dpid)
    def get_outport(self, src_node, dst_node):
        return self.outport_graph.get_outport(src_node, dst_node)
