__author__ = 'gabrielet'

import xml.etree.ElementTree as Xml

from model.vpn import Datapath, VirtualPrivateNetwork, Site, ProviderEdge
from model.nat import *

# For label generation
from utils.label import LabelGenerator


class Parser(object):
    def __init__(self):
        # The root of the XML document reporting the VPN's configuration
        self.root = None
        # Map of VPNs; map<name, model.vpn.VirtualPrivateNetwork>
        self.vpns = {}
        # Map of datapaths. map<dpid, model.vpn.Datapath>
        self.dps = {}
        # Map of provider edges. map<dpid, model.vpn.ProviderEdge>
        self.pes = {}
        # Map of natters. map<name, model.vpn.ProviderEdgeNat>
        self.nats = {}

        # Label generator
        self.label_generator = LabelGenerator.instance()

    def load(self, vpns_config_file):
        # Load file
        config = Xml.parse(vpns_config_file)
        self.root = config.getroot()

        # Load all datapaths
        for dp in self.root.findall('datapath'):
            name = dp.get('name')
            dpid = int(dp.get('dpid'))
            new_dp = Datapath(dpid, name)
            self.dps[dpid] = new_dp

        # Load all NATs
        for nat in self.root.findall('nat'):
            nat_name = nat.get('name')
            outport = nat.get('port')
            nat_label = self.label_generator.get_pe_label()
            # Create a new natter from a datapath
            dp = self._get_dp_from_name(nat_name)
            pe_nat = ProviderEdgeNat(dp.dpid, nat_name, nat_label, outport)
            # Add to the map
            self.nats[nat_name] = pe_nat
            self.dps[pe_nat.dpid] = pe_nat

        # Load all VPNs
        for vpn in self.root.findall('vpn'):
            # Create a new vpn
            name = vpn.get('name')
            vpn_label = self.label_generator.get_vpn_label()
            new_vpn = VirtualPrivateNetwork(name, vpn_label)

            # For each network, create a new pe (if not exists) and associate to it networks
            networks = vpn.findall('network')
            for network in networks:
                subnet = network.get('subnet')
                pe_name = network.get('pe')
                port = network.get('port')
                nat = network.get('nat')

                # Try to keep a PE from PEs map
                pe = self.pes.pop(pe_name, None)
                if pe is None:  # PE not found
                    # If PE was not found, then take it from datapaths map
                    dp = self._get_dp_from_name(pe_name)    # Take PE from datapaths maps
                    # Only here create a new PE; in this way, a duplicate PE is avoided
                    # Create a ProviderEdge instance
                    pe_label = self.label_generator.get_pe_label()
                    pe = ProviderEdge(dp.dpid, pe_name, pe_label)
                    # Add to the datapaths map
                    self.dps.pop(pe.dpid)
                    self.dps[pe.dpid] = pe

                # Create a site for this VPN
                site = Site(new_vpn, pe, port, subnet)
                # Add a NAT strategy
                # Check if this site is able to access to Internet
                if nat != "":
                    natter = self._create_nat_mapping(nat, site)
                    site.add_natter(natter)
                else:
                    site.add_natter(None)

                # Add this site to the VPN
                new_vpn.add_site(site)

                # Add PE to PEs' map
                self.pes[pe.name] = pe
            self.vpns[name] = new_vpn

    def _get_dp_from_name(self, name):
        dps = self.dps.values()
        for dp in dps:
            if name == dp.name:
                return dp

    def _create_nat_mapping(self, nat, site):
        # Take the nat from the map. If natter is not found, the return None
        natter = self.nats.pop(nat, None)
        # If natter is not none, create a mapping and strategy too; otherwise, return None
        if natter is not None:
            # Select all mapping with a certain VPN and PE
            nats = self.root.findall('nat')
            for n in nats:
                if n.get('name') == nat:
                    # Take all mapping
                    mappings = n.findall('mapping')
                    for m in mappings:
                        # Take correct mapping based on VPN name and PE name
                        if m.get('vpn') == site.vpn.name and m.get('pe') == site.pe.name:
                            m_type = m.get('type')
                            # Create the strategy for this mapping
                            factory_strategy = FactoryStrategy.instance()
                            strategy = factory_strategy.create_strategy(m_type)
                            # Add rule
                            rules = m.findall('rule')
                            for r in rules:
                                # Take ip and gr and add the rule to the strategy
                                ip = r.get('ip')
                                gr = r.get('gr')
                                strategy.add_rule(ip, gr)
                            # Create a mapping
                            mapping = Mapping(site, strategy)
                            natter.add_mapping(mapping)
                # Update natters' map
                self.nats[natter.name] = natter
        return natter

    def get_vpns(self):
        return self.vpns

    def get_dps(self):
        return self.dps

    def get_pes(self):
        return self.pes

    def get_nats(self):
        return self.nats


#if __name__ == "__main__":
#    p = Parser()
#    p.load()
